/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package sync4j.framework.engine.dm;

import java.net.URLEncoder;
import java.util.*;
import java.util.Map;

import sync4j.framework.core.*;
import sync4j.framework.protocol.ProtocolException;
import sync4j.framework.tools.Base64;
import sync4j.framework.tools.CommandIdGenerator;

/**
 * @author Stefano Fornari @ Funambol
 *
 * @version $Id: Util.java,v 1.1 2005/05/16 17:32:55 nichele Exp $
 */
public final class Util {

    // --------------------------------------------------------------- Constants

    private static final String MANAGEMENT_COMMANDS
        = ",Add,Alert,Copy,Delete,Exec,Get,Replace,Atomic,Sequence,";

    private static final String ALERT_OPTION_MINDT  = "MINDT" ;
    private static final String ALERT_OPTION_MAXDT  = "MAXDT" ;
    private static final String ALERT_OPTION_DR     = "DR"    ;
    private static final String ALERT_OPTION_MAXLEN = "MAXLEN";
    private static final String ALERT_OPTION_IT     = "IT"    ;
    private static final String ALERT_OPTION_ET     = "ET"    ;


    // --------------------------------------------------- Public static methods

    /**
     * Converts the given array of ManagementOperations in AbstractCommands.
     *
     * @param commands the ManagementOperations to convert
     * @param idGenerator generator for command ids
     *
     * @return the corresponding AbstractCommand array.
     */
    public static AbstractCommand[]
        managementOperations2commands(ManagementOperation[] operations ,
                                      CommandIdGenerator    idGenerator) {

            if ((operations == null) || (operations.length == 0)) {
                return new AbstractCommand[0];
            }

            AbstractCommand[] ret = new AbstractCommand[operations.length];
            for (int i=0; i<operations.length; ++i) {
                ret[i] = managementOperation2command(
                            operations[i], idGenerator.next(), idGenerator
                         );
            }

            return ret;
    }

    /**
     * Converts the given array of ManagementOperations in AbstractCommands.
     *
     * @param commands the ManagementOperations to convert
     * @param cmdId id for the new command
     *
     * @return the corresponding AbstractCommand array.
     */
    public static AbstractCommand
        managementOperation2command(ManagementOperation o, CmdID cmdId, CommandIdGenerator idGenerator) {

        if (o instanceof AtomicManagementOperation) {
            return new Atomic(
                cmdId,
                false,  // noResp
                null ,
                managementOperations2commands(((AtomicManagementOperation)o).getOperations(), idGenerator)
                );
        } else if (o instanceof SequenceManagementOperation) {
            return new Sequence(
                cmdId,
                false,  // noResp
                null ,
                managementOperations2commands(((SequenceManagementOperation)o).getOperations(), idGenerator)
                );
        } else if (o instanceof AddManagementOperation) {
            return new Add(
                cmdId,
                false,  // noResp
                null ,
                null ,
                nodes2Items(((AddManagementOperation)o).getNodes(), true)
            );
        } else if (o instanceof CopyManagementOperation) {
            // @TODO - Copy management operation to a Sequence command
            return null;
        } else if (o instanceof DeleteManagementOperation) {
            return new Delete(
                cmdId,
                false,  // noResp
                false,  // archive
                false,  // soft delete
                null ,
                null ,
                nodes2Items(((DeleteManagementOperation)o).getNodes(), false)
            );
        } else if (o instanceof ExecManagementOperation) {
            return new Exec(
                cmdId,
                false,  // noResp
                null ,
                nodes2Items(((ExecManagementOperation)o).getNodes(), true)[0]
            );
        } else if (o instanceof GetManagementOperation) {
            return new Get(
                cmdId,
                false,  // noResp
                null ,
                null ,
                null ,
                nodes2Items(((GetManagementOperation)o).getNodes(), false)
            );
        } else if (o instanceof ReplaceManagementOperation) {
            return new Replace(
                cmdId,
                false,  // noResp
                null ,
                null ,
                nodes2Items(((ReplaceManagementOperation)o).getNodes(), true)
            );
        } else if (o instanceof UserAlertManagementOperation) {
            UserAlertManagementOperation a = (UserAlertManagementOperation)o;

            return new Alert(
                cmdId,
                false,
                null,
                a.getAlertCode(),
                alert2Items(a)
            );
        }

        return null;
    }

    /**
     * Converts a mangement nodes map into an array of Item objects.
     * <p/>
     *
     * @param nodes the nodes map to convert
     * @param useValue should values be set in Items?
     *
     * @return the Item[]
     */
    public static Item[] nodes2Items(Map nodes, boolean useValue) {
        if ((nodes == null) || (nodes.size()==0)) {
            return new Item[0];
        }

        Item[] items = new Item[nodes.size()];

        Object key = null, value = null, treeNodeValue = null;

        String      format      = null;
        String      type        = null;
        Meta        meta        = null;
        ComplexData complexData = null;

        int j = 0;
        Iterator i = nodes.keySet().iterator();
        TreeNode treeNode = null;
        while(i.hasNext()) {
            key    = i.next();
            value  = nodes.get(key);
            meta   = null;
            format = null;
            type   = null;
            treeNodeValue = null;
            treeNode      = null;
            complexData   = null;

            if (!(value instanceof TreeNode)) {
                // create a dummy TreeNode
                value = new TreeNode((String)key, value);
            }

            treeNode = (TreeNode)value;

            format        = treeNode.getFormat();
            key           = treeNode.getName();
            treeNodeValue = treeNode.getValue();
            type          = treeNode.getType();

            if (format == null) {
                format = TreeNode.FORMAT_DEFAULT_VALUE;
            }

            if (format.equalsIgnoreCase(TreeNode.FORMAT_BINARY)) {
                if (treeNodeValue instanceof byte[]) {
                    complexData = new ComplexData(new String(Base64.encode( (byte[])treeNodeValue)));
                } else {
                    // get the byte[] of the string representation of the dataValue
                    complexData = new ComplexData(new String(Base64.encode(treeNodeValue.toString().
                        getBytes())));
                }
            } else if (format.equalsIgnoreCase(TreeNode.FORMAT_NODE)) {
                treeNodeValue = "";
            }

            if (useValue && (complexData == null)) {
                complexData = new ComplexData(String.valueOf(treeNodeValue));
            }

            meta = new Meta();
            meta.setFormat(format);
            meta.setType(type);

            items[j++] = new Item(
                            new Target(String.valueOf(key)),
                            null                           ,
                            (useValue) ? meta : null       ,
                            (useValue) ? complexData : null,
                            false   // moreData
                        );
        }

        return items;
    }

    /**
     * Converts an array of Status and Results commands into an array of
     * ManagementOperationResult objects.<br>
     * Only the status to the following commands are taken into account:
     *
     * <ul>
     *   <li>Add</li>
     *   <li>Copy</li>
     *   <li>Delete</li>
     *   <li>Exec</li>
     *   <li>Get</li>
     *   <li>Replace</li>
     *   <li>Atomic</li>
     *   <li>Sequence</li>
     * </ul>
     *
     * only if the their status code isn't contained in the given dischargedStatus.
     * <br>The <code>dischargedStatus</code> must be in the following format:<br>
     * <code>status1,status2,status3</code>. <br>This is used because for example the status
     * code 213 isn't a result code (or final code. The processor not must be notified
     * of this status).
     *
     * <p>
     * Note that in the case of a Status for a Get command, as per the specs
     * (OMA-SyncML-DMRepPro-V1_1_2-20030613-A) the following command is the
     * Results. Its content will be stored into the property nodes.
     * <p>
     * In all other cases, nodes will contain an entry for each Status' Item
     * (if there is any).
     *
     * @param commands the array of AbstractCommand to process
     * @param dischargedStatus comma separated list of status code to ignore
     *        (see method description)
     *
     * @return the corresponding ManagementOperationResult[]
     *
     * @throws ProtocolException in case of something wrong with the protocol
     */
    public static ManagementOperationResult[]
    operationResults(AbstractCommand[] commands, String dischargedStatus)
    throws ProtocolException {
        if ((commands == null) || (commands.length==0)) {
            return new ManagementOperationResult[0];
        }

        TreeMap results = new TreeMap(new CmdIdComparator());

        //
        // We first pass throgh all statuses and fill the results map.
        // Then we look for Results
        //

        //
        // NOTE: accordingly with the specs "The optional MsgRef element type
        // specifies the MsgID of the associated SyncML request. If the MsgRef
        // is not present in a Results element type, then the MsgRef value of
        // "1" MUST be assumed."
        //

        //
        // 1. Status processing
        //
        String key = null;
        Status status = null;
        ManagementOperationResult s = null;
        String statusCode = null;
        for(int i=0; i<commands.length; ++i) {
            if (!(commands[i] instanceof Status)) {
                continue;
            }

            status = (Status)commands[i];

            if (MANAGEMENT_COMMANDS.indexOf(status.getCmd()) < 0) {
                continue;
            }

            statusCode = status.getData().getData();
            if (dischargedStatus.indexOf(statusCode) != -1) {
                continue;
            }

            key = status.getMsgRef();
            key = ((key == null) ? "1" : key)
                + '-'
                + status.getCmdRef()
                + '-'
                + status.getStatusCode()
                ;

            s = (ManagementOperationResult)results.get(key);
            if (s == null) {
                String cmd = status.getCmd();
                s = new ManagementOperationResult();
                s.setStatusCode(status.getStatusCode());
                s.setCommand(cmd);

                if (Alert.COMMAND_NAME.equals(cmd)) {
                    s.setNodes(nodesFromItems(status));
                } else {
                    s.setNodes(nodesFromTargetRefs(status));
                }

                results.put(key, s);
            }
        }

        //
        // 2. Results processing
        //
        Results res = null;
        for(int i=0; i<commands.length; ++i) {
            if (!(commands[i] instanceof Results)) {
                continue;
            }

            res = (Results)commands[i];

            key = res.getMsgRef();
            key = ((key == null) ? "1" : key)
                + '-'
                + res.getCmdRef()
                + "-200"  // we suppose there was a 200 status, otherwise no
                          // Results should be here
                ;

            s = (ManagementOperationResult)results.get(key);

            if (s == null) {
                //
                // This should not happen!
                //
                throw new ProtocolException( "Results "
                                           + key
                                           + " is without corresponding Status"
                                           );
            }

            s.addNodes(nodesFromItems(res));
        }

        //
        // Converts results into a ManagementOperationResult[]
        //
        ManagementOperationResult[] ret
            = new ManagementOperationResult[results.size()];

        int i = 0;
        Iterator iter = results.values().iterator();
        while (iter.hasNext()) {
            ret[i++] = (ManagementOperationResult)iter.next();
        }

        return ret;
    }

    /**
     * Returns a Map from the Items of a ItemizedCommand. If the given command
     * is not a ItemizedCommand, an empty map is returned.
     *
     * @param cmd the status command
     *
     * @return a Map from the Items of a ItemizedCommand command
     */
    public static Map nodesFromItems(AbstractCommand cmd) {
        HashMap ret = new HashMap();

        if (!(cmd instanceof ItemizedCommand)) {
            return ret;
        }

        String node  = null;
        Object value = null;

        if (((ItemizedCommand)cmd).getItems() == null) {
            return ret;
        }

        int c = 0;
        Item item = null;
        Iterator i = ((ItemizedCommand)cmd).getItems().iterator();
        String format = null;
        Data itemData = null;
        while (i.hasNext()) {
            item  = (Item)i.next();
            format = getItemFormat(item);

            node = (item.getSource() == null)
                 ? String.valueOf(++c)
                 : item.getSource().getLocURI();

            itemData = item.getData();
            if (itemData != null) {
                value = itemData.getData();

                if (format.equalsIgnoreCase(TreeNode.FORMAT_DEFAULT_VALUE)) {
                    value = String.valueOf(value);
                } else if (format.equalsIgnoreCase(TreeNode.FORMAT_BINARY)) {
                    value = Base64.decode(String.valueOf(value).getBytes());
                } else if (format.equalsIgnoreCase(TreeNode.FORMAT_BOOL)) {
                    value = new Boolean((String)value);
                } else if (format.equalsIgnoreCase(TreeNode.FORMAT_INT)) {
                    try {
                        value = new Integer( (String)value);
                    } catch (NumberFormatException e) {
                        // if the node not contains a valid integer, adds an exception
                        value = new ManagementException(
                            "Node with format int not contains a valid integer value (" + value +
                            ")");
                    }
                }

                if (format.equalsIgnoreCase("node")) {
                    value = new TreeNode(item.getSource().getLocURI(), value, format);
                }
            }

            ret.put(node, (value == null) ? "" : value);
        }

        return ret;
    }

    /**
     * Returns a Map from the TargetRefs of a ResponseCommand. If the given
     * command is not a ResponseCommand, an empty map is returned.
     *
     * @param cmd the status command
     *
     * @return a Map from the TargetRefs of a ResponseCommand command
     */
    public static Map nodesFromTargetRefs(AbstractCommand cmd) {
        HashMap ret = new HashMap();

        if (!(cmd instanceof ResponseCommand)) {
            return ret;
        }

        String node = null;

        if (((ResponseCommand)cmd).getTargetRef() == null) {
            return ret;
        }

        TargetRef ref = null;
        Iterator i = ((ResponseCommand)cmd).getTargetRef().iterator();
        while (i.hasNext()) {
            ref  = (TargetRef)i.next();
            node  = ref.getValue();

            ret.put(node, "");
        }

        return ret;
    }

    /**
     * Returns an array of Item objects to be set into an Alert. The array is
     * at least two length long: the first Item is the option string, the others
     * are the real alerts.
     *
     * @param alert the UserAlertManagementObject instance
     *
     * @return the corresponding Item[] array
     */
    public static Item[] alert2Items(final UserAlertManagementOperation alert) {

        String[] alerts = alert.getAlerts();

        //
        // We have to create one item for the options string, plus one item for
        // each alert messages
        //
        Item[] items = new Item[(alerts == null) ? 1 : alerts.length+1];

        //
        // First of all, creates the option string.
        //
        StringBuffer options = new StringBuffer();

        int i = alert.getMinDisplayTime();
        if (i > 0) {
            options.append(ALERT_OPTION_MINDT).append('=').append(i).append('&');
        }
        i = alert.getMaxDisplayTime();
        if (i > 0) {
            options.append(ALERT_OPTION_MAXDT).append('=').append(i).append('&');
        }

        String s = alert.getDefaultResponse();
        if (s != null) {
            try {
                options.append(ALERT_OPTION_DR).append('=').append(URLEncoder.encode(s, "UTF-8")).append('&');
            } catch (java.io.UnsupportedEncodingException e) {
                options.append(e.getMessage());
            }
        }

        i = alert.getMaxLength();
        if (i > 0) {
            options.append(ALERT_OPTION_MAXLEN).append('=').append(i).append('&');
        }

        char c = alert.getInputType();
        if (c != ' ') {
            options.append(ALERT_OPTION_IT).append('=').append(c).append('&');
        }

        c = alert.getEchoType();
        if (c != ' ') {
            options.append(ALERT_OPTION_ET).append('=').append(c).append('&');
        }
        if (options.length() > 0) {
            options.deleteCharAt(options.length() - 1);
        }

        //
        // Note the here alert options will always terminate with an additional
        // '&' that must be removed.
        //
        items[0] = new Item(null, null, null, new ComplexData(options.toString()), false);

        //
        // Now we create a new Item for each alert message
        //
        for (i=1; i < items.length; ++i) {
            items[i] = new Item(null, null, null, new ComplexData(alerts[i-1]), false);
        }

        return items;
    }

    // --------------------------------------------------------- Private methods

    /**
     * Returns the format of the item. If the format is not specified, returns
     * the <code>ITEM_FORMAT_DEFAULT</code>
     *
     * @param item the item
     * @return the format of the given item
     */
    private static String getItemFormat(Item item) {
        Meta meta = item.getMeta();
        if (meta == null) {
            return TreeNode.FORMAT_DEFAULT_VALUE;
        }
        String format = meta.getFormat();
        if (format == null) {
            format = TreeNode.FORMAT_DEFAULT_VALUE;
        }

        return format;
    }

    // --------------------------------------------------------- Private classes

    /**
     * This class compares two strings in the form of keys used by the
     * <i>operationResults</i> methods. They are formatted as follows:<br>
     * <i>{msgref}-{cmdref}-{status}</i><br>
     * The comparison must be done on the <i>{cmdref}</i> only, considering it a
     * number.
     */
    private static class CmdIdComparator
    implements Comparator {
        /**
         * @see java.util.Comparator
         *
         * @param o1 first argument
         * @param o2 second argument
         *
         * @return a negative integer, zero, or a positive integer as the first
         * argument is less than, equal to, or greater than the second.
         */
        public int compare(Object o1, Object o2) {
            String value1 = null;
            String value2 = null;

            if (!((o1 instanceof String) && (o2 instanceof String))) {
                throw new IllegalArgumentException( "o1 ("
                                                  + o1
                                                  + ") and o2 ("
                                                  + o2
                                                  + ") must be string!"
                                                  );
            }

            value1 = (String)o1;
            value2 = (String)o2;

            return extractCmdId(value1) - extractCmdId(value2);
        }

        /**
         * Given a string in the form of <i>{msgref}-{cmdref}-{status}</i>,
         * this method extracts the {cmdref} and returns it as an int.
         * If the string is unparsable, IllegalArgumentException is thrown.
         *
         * @param s the string to be processed
         *
         * @return the cmdref as int
         *
         * @throws IllegalArgumentException if s is not in the form <i>{msgref}-{cmdref}-{status}</i>
         */
        private int extractCmdId(final String s)
        throws IllegalArgumentException {
            if (s == null) {
                throw new IllegalArgumentException("s cannot be null");
            }

            int p1 = s.indexOf('-');
            if ((p1 <= 0) || (p1 == s.length()-1)) {
                throw new IllegalArgumentException("s is not in the form {msgref}-{cmdref}-{status}");
            }

            int p2 = s.indexOf('-', p1+1);

            if (p2 <= 0) {
                throw new IllegalArgumentException("s is not in the form {msgref}-{cmdref}-{status}");
            }

            try {
                return Integer.parseInt(s.substring(p1+1, p2));
            } catch (NumberFormatException e) {
                throw new IllegalArgumentException("s is not in the form {msgref}-{cmdref}-{status}");
            }
        }
    }
}