/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


package sync4j.framework.security;

import java.util.HashMap;
import java.util.logging.Logger;
import java.util.logging.Level;

import java.security.Principal;
import javax.security.auth.*;
import javax.security.auth.login.*;
import javax.security.*;

import sync4j.framework.security.Officer;
import sync4j.framework.security.jaas.CredentialHandler;
import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.core.Cred;
import sync4j.framework.core.Authentication;
import sync4j.framework.config.Configuration;
import sync4j.framework.config.ConfigurationConstants;
import sync4j.framework.core.HMACAuthentication;
import sync4j.framework.core.Constants;
import sync4j.framework.tools.MD5;
import sync4j.framework.tools.Base64;
import sync4j.framework.server.store.*;
import sync4j.framework.server.Sync4jDevice;
import sync4j.framework.server.SyncUser;

import sync4j.server.admin.UserManager;


/**
 * This implementation of <i>Officier</i>
 *
 * @author  Stefano Nichele @ Funambol.com
 */
public class DBOfficer implements Officer, java.io.Serializable {

    // ---------------------------------------------------------- Protected data

    protected transient Logger log = Sync4jLogger.getLogger();
    protected PersistentStore ps = null;

    // ------------------------------------------------------------- Public data
    // -------------------------------------------------------------- Properties

    /**
     * Has the last login failed for incorrect login/password?
     */
    private boolean loginFailed = false;

    public boolean isLoginFailed() {
        return loginFailed;
    }

    /**
     * Has the last login failed for expired temporary login?
     */
    private boolean loginExpired = false;

    /**
     * Can use a guest credential if there is not credential in the message?
     */
    private boolean guestEnabled = true;

    public boolean isGuestEnabled() {
        return this.guestEnabled;
    }

    public void setGuestEnabled(boolean guestEnabled) {
        this.guestEnabled = guestEnabled;
    }

    /**
     * Which type of authetication does impose the server?
     */
    private String authType = "syncml:auth-md5";

    public String getAuthType() {
        return this.authType;
    }

    public void setAuthType(String authType) {
        this.authType = authType;

        // the authType is supported!!!!
        this.supportedAuthType = this.supportedAuthType + this.authType;
    }


    private String supportedAuthType = null;

    public String getSupportedAuthType() {
        return this.supportedAuthType;
    }

    public void setSupportedAuthType(String supportedAuthType) {

        // the authType is supported!!!!
        this.supportedAuthType = this.authType + supportedAuthType;
    }


    // ---------------------------------------------------------- Public methods

    /**
     * Authenticates a credential
     *
     * @param credential the credential to be authenticated
     *
     * @return true if the credential is autenticated, false otherwise
     */
    public boolean authenticate(Cred credential) {

        String type = credential.getType();

        if ( (Constants.AUTH_TYPE_BASIC).equals(type)  ) {
            return authenticateBasicCredential(credential);
        } else if ( (Constants.AUTH_TYPE_MD5).equals(type)  ) {
            return authenticateMD5Credential(credential);
        } else if ( (Constants.AUTH_TYPE_HMAC).equals(type)  ) {
            return authenticateHMACCredential(credential);
        } else {
            return false;
        }
    }

    /**
     * Authorizes a resource.
     *
     * @param principal the requesting entity
     * @param resource the name (or the identifier) of the resource to be authorized
     *
     * @return true if the credential is authorized to access the resource, false
     *         otherwise
     */
    public boolean authorize(Principal principal, String resource) {
        return true;
    }

    /** Un-authenticates a credential.
     *
     * @param credential the credential to be unauthenticated
     */
    public void unAuthenticate(Cred credential) {
        //
        // Do nothing. In the current implementation, the authentication is
        // discarde as soon as the LoginContext is garbage collected.
        //
    }

    /**
     * @see sync4j.framework.security.Officer
     */
    public boolean isAccountExpired() {
        return loginExpired;
    }


    // --------------------------------------------------------- Private Methods
    private void readObject(java.io.ObjectInputStream in)
    throws java.io.IOException, ClassNotFoundException {
        in.defaultReadObject();
        log = Sync4jLogger.getLogger();
    }


    private boolean authenticateBasicCredential(Cred credential) {
        SyncUser       user   = null                          ;
        Authentication auth   = credential.getAuthentication();

        String username = null;
        String password = null;

        String s = new String(Base64.decode(credential.getData()));

        int p = s.indexOf(':');

        if (p == -1) {
            username = s;
            password = "";
        } else {
            username = (p>0) ? s.substring(0, p) : "";
            password = (p == (s.length()-1)) ? "" : s.substring(p+1);
        }

        if (log.isLoggable(Level.FINE)) {
            log.fine("Username: " + username + "; password: " + password);
        }

        try {
            user = readUser(username);
        } catch (NotFoundException e) {
            if (log.isLoggable(Level.FINE)) {
                log.fine("User " + username + " not found!");
            }
        } catch (PersistentStoreException e) {
            log.severe( "Error reading the user "
                      + username
                      + ": "
                      + e.getMessage()
                      );
            log.throwing(getClass().getName(), "authenticateBasicCredential", e);

            return false;
        }

        return password.equals(user.getPassword());
    }

    private boolean authenticateMD5Credential(Cred credential) {

        Sync4jDevice   device = null                          ;
        Authentication auth   = credential.getAuthentication();

        try {
            device = readDevice(auth.getDeviceId());
        } catch (PersistentStoreException e) {
            log.severe("Error configuring the persistent store: " + e);
            log.throwing(getClass().getName(), "authenticateMD5Credential", e);

            return false;
        }

        //
        // NOTE: the digest read from the store is formatted as
        // b64(H(username:password))
        //
        String userDigest  = device.getDigest();
        byte[] clientNonce = auth.getNextNonce().getValue();

        if (log.isLoggable(Level.FINE)) {
            log.fine("userDigest: " + userDigest);
        }

        byte[] userDigestBytes = userDigest.getBytes();

        //
        // computation of the MD5 digest
        //
        // Creates a unique buffer containing the bytes to digest
        //
        byte[] buf = new byte[userDigestBytes.length + 1 + clientNonce.length];

        System.arraycopy(userDigestBytes, 0, buf, 0, userDigestBytes.length);
        buf[userDigestBytes.length] = (byte)':';
        System.arraycopy(clientNonce, 0, buf, userDigestBytes.length+1, clientNonce.length);

        byte[] digest = MD5.digest(buf);

        //
        // encoding digest in Base64 for comparation with digest sent by client
        //
        String serverDigestNonceB64 = new String(Base64.encode(digest));

        //
        // digest sent by client in format b64
        //
        String msgDigestNonceB64 = auth.getData();

        if (log.isLoggable(Level.FINE)) {
            log.fine("serverDigestNonceB64: " + serverDigestNonceB64);
            log.fine("msgDigestNonceB64: "    + msgDigestNonceB64   );
        }

        if (!msgDigestNonceB64.equals(serverDigestNonceB64)) {
            return false;
        }

        auth.setUsername(userDigest);

        return true;

    }

    private boolean authenticateHMACCredential(Cred credential) {
        Sync4jDevice   device = null                          ;
        Authentication auth   = credential.getAuthentication();

        if ( !(auth instanceof HMACAuthentication) ) {
            throw new IllegalStateException("Authentication not valid!");
        }

        try {
            device = readDevice(auth.getDeviceId());
        } catch (PersistentStoreException e) {
            log.severe("Error configuring the persistent store: " + e);
            log.throwing(getClass().getName(), "authenticateMD5Credential", e);

            return false;
        }

        String userMac       = ((HMACAuthentication)auth).getUserMac()      ;
        String calculatedMac = ((HMACAuthentication)auth).getCalculatedMac();

        if ((userMac != null) && (userMac.equals(calculatedMac))) {
            auth.setUsername(device.getDigest());
            return true;
        }

        return false;
    }

    /**
     * Reads a device from the persistent store from the device id.
     *
     * @param deviceId the device id
     *
     * @return a Sync4jDevice representing the device
     *
     * @throws PersistentStoreException in case of errors reading from the db
     */
    private Sync4jDevice readDevice(String deviceId)
    throws PersistentStoreException {
        Configuration config = Configuration.getConfiguration();

        PersistentStore ps =
            (PersistentStore)config.getBeanInstance(ConfigurationConstants.CFG_PERSISTENT_STORE);

        Sync4jDevice device = new Sync4jDevice(deviceId, null, null);
        ps.read(device);

        return device;
    }

    /**
     * Reads a user from the database store given his/her username.
     *
     * @param username the user name
     *
     * @throws NotFoundException if no users are found
     *         PersistentStoreException in case of persistent store errors
     */
    private SyncUser readUser(String username)
    throws NotFoundException, PersistentStoreException {
        Configuration config = Configuration.getConfiguration();

        UserManager um =
            (UserManager)config.getBeanInstance(ConfigurationConstants.CFG_USER_MANAGER);

        SyncUser[] results = um.getUsers(
            new WhereClause("username", new String[] { username }, WhereClause.OPT_EQ, false)
        );

        if ((results == null) || (results.length == 0)) {
            throw new NotFoundException(username);
        }

        return results[0];
    }


}