/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package sync4j.server.notification;

import java.io.BufferedOutputStream;
import java.net.Socket;
import java.net.URL;
import java.util.logging.Level;
import java.util.logging.Logger;

import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.notification.NotificationConstants;
import sync4j.framework.notification.NotificationException;
import sync4j.framework.notification.Sender;
import sync4j.framework.logging.Sync4jLoggerName;


/**
 * This sender sends notification message via http (to use with SCTS DM)
 * <p/>
 * To obtain the same result of the SCTS server you must to set:
 * <ui>SessionId: 12345</ui>
 * <ui>ServerId: Scts</ui>
 *  <ui>ServerPwd: SimpleClient</ui>
 * <ui>ServerNonce: AOVuAA==   (this is already in base 64 format. Corresponds to 00E56E00)</ui>
 * <br/>and in <code>sync4j.server.notification.NotificationEngineImpl</code> you must to set:
 *  <ui>SYNCML_DM_PROTOCOL_VERSION = 1.1f</ui>
 *
 * @author Stefano Nichele @ Funambol
 * @version $Id: HttpSender.java,v 1.1 2005/05/16 17:32:56 nichele Exp $
 */
public class HttpSender implements Sender {

    // -------------------------------------------------------------- Constants
    // -------------------------------------------------------------- Properties
    private String deviceAddress = null;

    // ------------------------------------------------------------ Private data
    private Logger log = null;

    // --------------------------------------------------------------- Constructor
    public HttpSender() {
        log = Sync4jLogger.getLogger(Sync4jLoggerName.SERVER_DM_NOTIFICATION);
    }

    // ------------------------------------------------------------ Public Methods

    /**
     * Gets the address device
     *
     * @return  the address device
     */
    public String getDeviceAddress() {
        return deviceAddress;
    }

    /**
     * Sets the address device
     *
     * @param  the address device
     */
    public void setDeviceAddress(String deviceAddress) {
        this.deviceAddress = deviceAddress;
    }

    /**
     * Send the given message via http
     * @param messageType int
     * @param sessionId String
     * @param phoneNumber String
     * @param message byte[]
     */
    public void sendMessage(int messageType, String phoneNumber, byte[] message, String info)
        throws NotificationException {

        if (log.isLoggable(Level.INFO)) {
            log.info("sendMessage: " + message + " (length: " + message.length + ")");
        }


        if (messageType != NotificationConstants.MESSAGE_TYPE_NOTIFICATION_GENERIC) {
            throw new NotificationException("This sender is usable only for notification message");
        }

        byte[] nestedRequest = buildNestedRequest(message);;
        byte[] request = buildRequest(nestedRequest);

        try {
            sendHttpRequest(request);
        } catch (Exception e) {
            throw new NotificationException("Error sending the message via http", e);
        }

    }

    /**
     * Sends more messages
     * @param messageType the type of the message. See {@link NotificationConstants}
     * @param contentType the contentType
     * @param macs the macs list
     * @param authMethods the authentication methods
     * @param phoneNumbers the phone number list
     * @param messages the messages to send
     * @param info application specific info
     * @throws NotificationException
     */
    public void sendMessages(int      messageType,
                             String   contentType,
                             String[] digest,
                             int[]    authMethods,
                             String[] phoneNumbers,
                             byte[][] messages,
                             String info) throws NotificationException {

        if (log.isLoggable(Level.INFO)) {
            log.info("sendMessage with empty implementation");
        }
    }

    /**
     * Sends messages
     * @param messageType the type of the message. See {@link NotificationConstants}
     * @param phoneNumbers the phone numbers list of the target device
     * @param messages the messages to send
     * @param info application specific info
     * @throws NotificationException
     */
    public void sendMessages(int      messageType,
                             String[] phoneNumbers,
                             byte[][] messages,
                             String   info) throws NotificationException {

        for (int i=0; i<phoneNumbers.length; i++) {

            //
            // For the first message not wait
            //
            if (i != 0) {
                System.out.println("Wait 20 s for test with SCTS");
                try {
                    Thread.sleep(20000);
                } catch (InterruptedException ex) {
                }

            }
            sendMessage(messageType, phoneNumbers[i], messages[i], info);
        }
    }


    private void sendHttpRequest(byte[] message) throws Exception {
        if (log.isLoggable(Level.INFO)) {
            log.info("Send message to: " + deviceAddress);
        }
        URL url = new URL(deviceAddress);
        String host = url.getHost();
        int port = url.getPort();

        Socket conn = new Socket(host, port);

        // Send data
        BufferedOutputStream wr = new BufferedOutputStream(conn.getOutputStream());

        wr.write(message);
        wr.flush();

        wr.close();
    }

    private byte[] buildRequest(byte[] nestedRequest) {
        StringBuffer sb = new StringBuffer("POST /wappush HTTP/1.1\r\n");
        sb.append("Host: 127.0.0.1\r\n");
        sb.append("Content-Type: application/http\r\n");
        sb.append("Content-Length: ").append(nestedRequest.length).append("\r\n");
        sb.append("X-Wap-Push-OTA-Version: 1.0\r\n\r\n");

        sb.append(new String(nestedRequest));

        return sb.toString().getBytes();

    }


    private byte[] buildNestedRequest(byte[] message) {
        StringBuffer sb = new StringBuffer("HTTP/1.1 200 OK\r\n");
        sb.append("Content-Language: en\r\n");
        sb.append("Content-Length: ").append(message.length).append("\r\n");
        sb.append("Content-Type: application/vnd.syncml.notification\r\n\r\n");

        byte[] header = sb.toString().getBytes();

        byte[] nestedRequest = new byte[header.length + message.length];

        System.arraycopy(header, 0, nestedRequest, 0, header.length);
        System.arraycopy(message, 0, nestedRequest, header.length, message.length);

        return nestedRequest;


    }

}