/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package sync4j.transport.http.server;

import java.io.File;
import java.io.IOException;

import java.net.URL;
import java.net.MalformedURLException;

import java.util.logging.Logger;
import java.util.logging.Level;

import javax.naming.InitialContext;

import sync4j.framework.transport.http.SyncHolder;

import sync4j.framework.core.Sync4jException;
import sync4j.framework.config.*;
import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.server.SyncResponse;
import sync4j.framework.server.error.ServerException;
import sync4j.framework.server.error.ServerFailureException;
import sync4j.framework.server.error.NotImplementedException;
import sync4j.framework.server.SyncResponse;
import sync4j.framework.protocol.ProtocolException;

import sync4j.server.engine.SyncAdapter;

/**
 * Implementes a <i>SyncHolder</i>.
 *
 * @author  Luigia Fassina @ Funambol
 * @version $Id: LocalSyncHolder.java,v 1.1 2005/05/16 17:32:57 nichele Exp $
 */
public class LocalSyncHolder implements SyncHolder {

    // --------------------------------------------------------------- Constants

    private Logger log = Sync4jLogger.getLogger("transport.http");
    private static final String ENV_SERVER_CONFIG_URI
        = "java:comp/env/server/config_uri";
    private static final String ENV_SERVER_CONFIG_PATH
        = "java:comp/env/server/config_path";

    // ------------------------------------------------------------ Private data

    private SyncAdapter syncAdapter       = null;
    private long        creationTimestamp       ;

    // ------------------------------------------------------------ Constructors
    public LocalSyncHolder() throws ServerException {
        syncAdapter = new SyncAdapter(loadConfiguration());
        creationTimestamp = System.currentTimeMillis();
    }

    // ---------------------------------------------------------- Public methods

    /** Processes an incoming message.
     *
     * @param requestData the SyncML request as stream of bytes
     * @param contentType the content type associated with the request
     * @param hmacHeader the hmac value associated with the request
     *
     * @return the SyncML response as a <i>ISyncResponse</i> object
     *
     * @throws ServerException in case of a server error
     *
     */
    public SyncResponse processMessage(byte[] requestData, String contentType, String hmacHeader)
    throws ServerException {
        return syncAdapter.processMessage(requestData, contentType, hmacHeader);
     }

    public void setSessionId(String sessionId) throws Sync4jException {
        syncAdapter.setSessionId(sessionId);
    }

    public String getSessionId() {
        return syncAdapter.getSessionId();
    }

    /** Called when the SyncHolder is not required any more. It gives the holder
     * an opportunity to do clean up and releaseing of resources.
     *
     * @throws java.lang.Exception in case of error. The real exception is stored
     * in the cause.
     *
     */
    public void close() throws Exception {
        syncAdapter.endSync();
    }

    /**
     * Returns the creation timestamp (in milliseconds since midnight, January
     * 1, 1970 UTC).
     *
     * @see sync4j.framework.transport.http.SyncHolder
     */
    public long getCreationTimestamp() {
        return creationTimestamp;
    }

    // --------------------------------------------------------- Private methods

    /**
     * Loads the configuration for the server. The URI from which the configuration
     * is must be set as the envirnoment property named as <i>ENV_SERVER_CONFIG_URI</i>.
     *
     * @throws ConfigurationException in case of errors.
     */
    private Configuration loadConfiguration()
    throws ConfigurationException {
        URL configURI = null, configPath = null;
        Configuration config = Configuration.getConfiguration();

        try {
            InitialContext ctx = new InitialContext();

            //
            // The following two entries should be two URIs, but in the case
            // they are not (they do not start with http://, https:// or file://,
            // they are considered simple files
            //
            configURI  = fixURI((String)ctx.lookup(ENV_SERVER_CONFIG_URI ));
            configPath = fixURI((String)ctx.lookup(ENV_SERVER_CONFIG_PATH));

            config.setClassLoader(
                new ConfigClassLoader(
                    new URL[] { configPath },
                    getClass().getClassLoader()
                )
            );

            config.load(configURI.toString());

            if (log.isLoggable(Level.FINEST)) {
                log.finest("Configuration: " + config);
            }

            return config;
        } catch (ConfigurationException e) {
            e.printStackTrace();
            throw e;
        } catch (Exception e) {
            throw new ConfigurationException("Error loading configuration from " + configURI, e);
        }
    }

    /**
     * Checks if the given string is an accepted URL (starting with http://,
     * https:// or file://). If yes, a new URL object representing the given
     * url is returned; otherwise, the given string is considered a file name
     * and a new URL is obtained calling File.toURL().
     *
     * @param s the string to check
     *
     * @return the corresponding URL if the string represents a URL or the
     *         fixed URL if the string is a pathname/filename
     *
     * @throws MalformedURLException
     */
    private URL fixURI(final String s)
    throws MalformedURLException {
        String minS = s.toLowerCase();

        try {
            return new URL(s);
        } catch (MalformedURLException e) {
            //
            // This is not a URL, let's consider it just a file
        }

        try {
            return new File(new File(s).getCanonicalPath()).toURL();
        } catch (IOException e) {
            throw new MalformedURLException("Unable to convert" + s + " to a URL");
        }
    }

}